local M = {}

local imguiUtils = require('ui/imguiUtils')
local im = ui_imgui

local logTag = "infoBox"

M.state = "hidden"

local ready = im.BoolPtr(true)
local requeue = false
local baseColor = im.ImVec4(1, 1, 1, 1)
local queue, styles, mainData = {}, {}, {}
local timers = {
  queue = 0,
  resetS = 0,
  resetA = 0
}
local elements = {
  text = 0,
  windowBase = 2,
  windowBorder = 5,
  comboBase = 7,
  comboHover = 8,
  titleBarInactive = 10,
  titleBarActive = 11,
  buttonBase = 21,
  buttonHover = 22,
  buttonPress = 23,
  collapserBase = 24,
  collapserHover = 25,
  collapserPress = 26,
  separator = 27,
  resizerBase = 30,
  resizerHover = 31,
  resizerPress = 32
}
local options = {
  hoverColor = nil,
  colorValue = "Default",
  difficulty = "Normal"
}

local function timeToString(t) -- returns a user friendly time string
  local minutes = math.floor(t / 60);
  local seconds = t - (minutes * 60);
  return string.format("%02.0f:%05.2f", minutes, seconds)
end

local function parseColor(old) -- parses a string color to an ImVec4
  local new = old and stringToTable(old, " ") or nil
  if not new or #new ~= 4 then
    log("I", logTag, "Invalid color string! Using default color...")
    return baseColor
  else
    return im.ImVec4(tonumber(new[1]), tonumber(new[2]), tonumber(new[3]), tonumber(new[4]))
  end
end

local function setStyles(data) -- sets various style data for the imgui window
  if not data then data = {} end
  table.clear(styles)
  styles.elements = {}
  styles.images = {}
  styles.baseWidth = data.baseWidth or 400
  styles.baseHeight = data.baseHeight or 200
  styles.fontSize = data.fontSize or 1
  styles.color1 = parseColor(data.color1)
  styles.color2 = parseColor(data.color2)
  styles.color3 = parseColor(data.color3)
  if data.elements then
    for k, v in pairs(data.elements) do
      if elements[k] and v then
        styles.elements[elements[k]] = parseColor(v)
      end
    end
  end
  if data.images then
    for k, v in pairs(data.images) do
      styles.images[k] = imguiUtils.texObj(v)
    end
  end
end

local function resizeBox(data) -- returns the expected dimensions of the text box
  if not data or data.noResize then return styles.baseWidth, styles.baseHeight end
  local width, height, bestWidth = 0, 0, 0
  local size = styles.fontSize
  for i, v in ipairs(data.text) do
    local lineWidth = string.len(tostring(v)) * 7 * size
    if lineWidth >= bestWidth then
      bestWidth = lineWidth
    end
  end
  width = bestWidth
  if data.image then
    width = math.max(260, width)
    height = height + 40
  end
  if data.title then height = height + 28 * size end
  if data.subtitle then height = height + 28 * size end
  for i, v in ipairs(data.text) do
    height = height + 28 * size
  end
  width = math.ceil(clamp(width, 80, 1280))
  height = math.ceil(clamp(height, 20, 720))
  return width, height
end

local function showInfoBox(data) -- shows a pop-up UI in the corner
  if not data.width then data.width = styles.baseWidth end
  if not data.height then data.height = styles.baseHeight end
  im.SetNextWindowPos(im.ImVec2(0, 60), im.Cond_Always)
  im.SetNextWindowSize(im.ImVec2(data.width, data.height), im.Cond_Always)
  for k, v in pairs(styles.elements) do
    im.PushStyleColor2(k, v)
  end
  if im.Begin("Info Box ", ready, im.flags(im.WindowFlags_NoTitleBar, im.WindowFlags_NoResize, im.WindowFlags_NoScrollbar)) then
    im.SetWindowFontScale(styles.fontSize)
    if data.image and styles.images[data.image] then
      im.Image(styles.images[data.image].texId, im.ImVec2(256, 24), im.ImVec2(0, 0), im.ImVec2(1, 1), im.GetStyleColorVec4(im.Col_Text))
    end
    if data.title then
      im.TextColored(styles.color1, data.title)
    end
    if data.subtitle then
      im.TextColored(styles.color2, data.subtitle)
    end
    for i, v in ipairs(data.text) do
      im.Text(v)
    end
    im.End()
  end
  for k, v in pairs(styles.elements) do
    im.PopStyleColor()
  end
end

local function showMainBox(data) -- shows a main window
  for k, v in pairs(styles.elements) do
    im.PushStyleColor2(k, v)
  end
  if M.state == "mini" then
    im.SetNextWindowPos(im.ImVec2(24, 24), im.Cond_Always)
    im.SetNextWindowSize(im.ImVec2(50, 50), im.Cond_FirstUseEver)
    if im.Begin("Race Extras (Mini) ", ready, im.flags(im.WindowFlags_NoTitleBar, im.WindowFlags_NoResize, im.WindowFlags_NoScrollbar)) then
      im.SetWindowFontScale(styles.fontSize)
      if im.Button(" + ") then
        M.state = "full"
      end
      im.End()
    end
  elseif M.state == "full" then
    im.SetNextWindowPos(im.ImVec2(24, 24), im.Cond_Always)
    im.SetNextWindowSize(im.ImVec2(data.width, data.height), im.Cond_FirstUseEver)
    if im.Begin("Race Extras ", ready, im.flags(im.WindowFlags_NoTitleBar)) then
      im.SetWindowFontScale(styles.fontSize)
      if data.image and styles.images[data.image] then
        im.Image(styles.images[data.image].texId, im.ImVec2(256, 24), im.ImVec2(0, 0), im.ImVec2(1, 1), im.GetStyleColorVec4(im.Col_Text))
      end
      if data.title then
        im.TextColored(styles.color1, data.title)
      end
      if data.subtitle then
        im.TextColored(styles.color2, data.subtitle)
      end
      if data.text then
        im.TextWrapped(data.text)
      end
      if im.Button("Minimize") then
        M.state = "mini"
      end
      im.Separator()
      im.TextColored(styles.color3, "Wins: ")
      im.SameLine()
      im.Text(tostring(data.wins))
      im.TextColored(styles.color3, "Finishes: ")
      im.SameLine()
      im.Text(tostring(data.finishes))
      im.TextColored(styles.color3, "Money: ")
      im.SameLine()
      im.Text("$"..data.money)
      if im.CollapsingHeader1("Settings ") then
        im.PushItemWidth(160)
        if im.BeginCombo("Car Color ", options.colorValue) then
          for i, v in ipairs(data.colorsList) do
            local selected = (options.colorValue == v) and true or false
            if im.Selectable1(v, selected) then
              options.colorValue = v
              extensions.hook("onRaceExtrasAction", {playerColor = data.playerColors[v]})
            end
            if im.IsItemHovered() then
              options.hoverColor = data.playerColors[v]
            end
            if selected then
              im.SetItemDefaultFocus()
            end
          end
          im.EndCombo()
        else
          options.hoverColor = nil
        end
        im.PopItemWidth()
        if options.hoverColor then
          im.SameLine()
          local colorBox = im.ImVec4ToFloatPtr(parseColor(options.hoverColor))
          im.ColorEdit4("newColor", colorBox, im.flags(im.ColorEditFlags_NoInputs, im.ColorEditFlags_NoLabel))
        end
        im.PushItemWidth(160)
        if im.BeginCombo("AI Difficulty ", options.difficulty) then
          for i, v in ipairs(data.difficultyList) do
            local selected = (options.difficulty == v) and true or false
            if im.Selectable1(v, selected) then
              options.difficulty = v
              extensions.hook("onRaceExtrasAction", {difficulty = v})
            end
            if selected then
              im.SetItemDefaultFocus()
            end
          end
          im.EndCombo()
        end
        im.PopItemWidth()
      end
      if im.CollapsingHeader1("Stats ") then
        if data.racesList and data.racesList[1] then
          local columnWidth = math.floor((im.GetContentRegionAvailWidth() - 240) / 3)
          im.Columns(4, "Races Data", false)
          im.SetColumnWidth(0, 240)
          im.SetColumnWidth(1, columnWidth)
          im.SetColumnWidth(2, columnWidth)
          im.SetColumnWidth(3, columnWidth)
          im.TextColored(styles.color3, "Race Name")
          im.NextColumn()
          im.TextColored(styles.color3, "Wins")
          im.NextColumn()
          im.TextColored(styles.color3, "Finishes")
          im.NextColumn()
          im.TextColored(styles.color3, "Time")
          im.NextColumn()
          for i, v in ipairs(data.racesList) do
            local d = data.races[v]
            im.Text(d.name)
            im.NextColumn()
            im.Text(tostring(d.wins))
            im.NextColumn()
            im.Text(tostring(d.finishes))
            im.NextColumn()
            im.Text(timeToString(d.time))
            im.NextColumn()
          end
          im.Columns(1)
        end
        if data.minScore > 0 then
          im.Separator()
          im.TextColored(styles.color3, "Sum of Best Times: ")
          im.SameLine()
          im.Text(timeToString(data.time))
        end
      end
      if im.CollapsingHeader1("Achievements ") then
        if data.useAchievements then
          local done, notDone = "", ""
          local doneLength, notDoneLength = #data.achievementsDoneList, #data.achievementsNotDoneList
          for i, v in ipairs(data.achievementsDoneList) do
            done = done..data.achievementsRef[v].title
            if i < doneLength then
              done = done..", "
            end
          end
          for i, v in ipairs(data.achievementsNotDoneList) do
            notDone = notDone..data.achievementsRef[v].title
            if i < notDoneLength then
              notDone = notDone..", "
            end
          end
          im.TextColored(styles.color3, "Done: ")
          im.SameLine()
          im.TextWrapped(done)
          im.TextColored(styles.color3, "Not Done: ")
          im.SameLine()
          im.TextWrapped(notDone)
          im.TextColored(styles.color3, "Progress: ")
          im.SameLine()
          im.Text(doneLength.." / "..data.achievements.total)
        end
      end
      if im.CollapsingHeader1("Saved Data ") then
        if im.Button("Reset Stats ") then
          extensions.hook("onRaceExtrasAction", {resetS = true})
          timers.resetS = 5
        end
        if timers.resetS > 0 then
          im.SameLine()
          im.Text("Reset! Please restart scenario.")
        end
        if im.Button("Reset Achievements ") then
          extensions.hook("onRaceExtrasAction", {resetA = true})
          timers.resetA = 5
        end
        if timers.resetA > 0 then
          im.SameLine()
          im.Text("Reset! Please restart scenario.")
        end
      end
    im.End()
    end
  end
  for k, v in pairs(styles.elements) do
    im.PopStyleColor()
  end
end

local function queueInfo(data)
  if not data then return end
  if not data.time then data.time = 1e6 end
  if type(data.text) ~= "table" then
    data.text = {data.text}
  end
  if not data.width or not data.height then
    data.width, data.height = resizeBox(data)
  end
  if timers.queue == 0 then
    timers.queue = data.time
  end
  local new = data
  table.insert(queue, new)
end

local function stopInfo()
  table.clear(queue)
end

local function setupMainBox(data, mode)
  if not data then data = {} end
  stopInfo()
  M.state = mode
  if mode ~= "hidden" then
    if not data.width or data.height then
      data.width, data.height = 540, 480
    end
    mainData = data
  end
end

local function onExtensionLoaded()
  setStyles()
end

local function onUpdate(dt)
  local q = queue[1]
  if q then
    if timers.queue > 0.25 then
      showInfoBox(q)
    elseif timers.queue <= 0 then
      requeue = true
    end
  end
  if requeue and q then
    timers.queue = queue[2] and queue[2].time or 0
    table.remove(queue, 1)
    requeue = false
  end
  if M.state ~= "hidden" then
    showMainBox(mainData)
  end
  for k, v in pairs(timers) do
    timers[k] = v > 0 and v - dt or 0
  end
end

M.setStyles = setStyles
M.queueInfo = queueInfo
M.stopInfo = stopInfo
M.setupMainBox = setupMainBox
M.onUpdate = onUpdate
M.onExtensionLoaded = onExtensionLoaded

return M