local M = {}

local scenario = require("scenario/scenarios")
local scoreStats = require("statistics/statistics")

local player, racers, racePath, stats, finishers = {}, {}, {}, {}, {}
local mapNodes, mapGp, aiLua, finishPos, finishVec, finishRadius, watchNodeNames
local playerObj = be:getPlayerVehicle(0)
local firstLoad = true
local mainTimer = 0
local spawnDelay = -1
local aggroScale = 1
local raceLength = 0
local up = vec3(0, 0, 1)
local logTag = "raceExtras"
local carLimit = Engine.Platform.getCPUInfo().coresPhysical or 4
local userCarLimit = core_settings_settings.getValue("trafficAmount")
local ranks = {"1st", "2nd", "3rd", "4th", "5th", "6th", "7th", "8th", "9th", "10th"}
local series = {
  name = "default",
  saveFile = "settings/raceExtrasSave.json",
  achievementsFile = "",
  races = {},
  messages = {
    win = "You won the race!",
    finish = "won the race.",
    place = "You placed",
    finishTime = "Time:",
    finishReward = "Reward:",
    money = "Total cash:",
    achievements = "Achievements:",
    congrats = "Perfect!",
    stuck = "You got stuck!",
    flip = "You flipped over!"
  }
}
local event = {
  seriesFile = "",
  useMainBox = true,
  aiLua = "ai",
  damageCutoff = 10000,
  stuckTime = 3,
  crawlSpeed = 3,
  aggroValue = 1,
  aggroVariance = 0,
  aggroTurnWatch = 0,
  aggroTurnFactor = 0,
  aggroTurnCutoff = 0.7,
  aggroReboundRate = 0.05,
  aggroDownhill = 0.5,
  aggroRubberband = 0,
  aggroRubberbandBounds = {-0.4, 0.05},
  avoidCars = false,
  playerName = "Player",
  racerNames = {},
  rewards = {500, 250, 125, 60, 40, 20}
}
local spawn = {
  group = nil,
  quantity = 1,
  order = true,
  mode = "raceGridAlt",
  gap = 15
}

local function timeToString(t) -- returns a user friendly time string
  local minutes = math.floor(t / 60);
  local seconds = t - (minutes * 60);
  return string.format("%02.0f:%05.2f", minutes, seconds)
end

local function setupSaveFile(data, flags) -- creates, modifies, and returns save data
  if not flags then flags = {} end
  local json = data or {
    level = event.level,
    series = series.name,
    races = {},
    achievements = {},
    useAchievements = true
  }
  if flags.stats or not next(json.races) then
    json.money = 0
    json.wins = 0
    json.finishes = 0
    json.score = 0
    json.minScore = 0
    json.time = 0
    for i, v in ipairs(series.races) do
      json.races[v] = {name = "race", wins = 0, finishes = 0, score = 0, time = 0}
      local scenarioFile = jsonReadFile(event.level.."scenarios/"..v..".json")
      json.races[v].name = scenarioFile[1].name
    end
  end
  if flags.achievements or not next(json.achievements) then
    local goals = jsonReadFile(event.achievementsFile)
    json.achievements = {
      notDone = {},
      done = {}
    }
    if goals then
      for k, v in pairs(goals) do
        table.insert(json.achievements.notDone, k)
      end
      json.achievements.total = #json.achievements.notDone
    else
      json.useAchievements = false
    end
  end
  jsonWriteFile(event.saveFile, json)
  return json
end

local function modifySaveFile(flags) -- recalls setupSaveFile to modify or reset data
  local json = jsonReadFile(series.saveFile)
  setupSaveFile(json, flags)
end

local function setCarColor(id, color) -- sets a car's main color
  local v = scenetree.findObjectById(id)
  if not v or not color then return end
  local color1, color2 = tostring(v.obj.color), tostring(v.obj.colorPalette0)
  local colorTable = stringToTable(color, " ")
  if colorTable[4] then
    local newColor = ColorF(colorTable[1], colorTable[2], colorTable[3], colorTable[4]):asLinear4F()
    v.obj.color = newColor
    if color1 == color2 then
      v.obj.colorPalette0 = newColor
    end
  end
end

local function getCondition(cond) -- returns true if an achievement condition statement checks out
  local valid = false
  local var1 = type(cond[1]) == "number" and cond[1] or stats[cond[1]]
  local var2 = type(cond[3]) == "number" and cond[3] or stats[cond[3]]
  if var1 and var2 then
    if cond[2] == "lte" then
      if var1 <= var2 then
        valid = true
      end
    elseif cond[2] == "gte" then
      if var1 >= var2 then
        valid = true
      end
    else
      if var1 == var2 then
        valid = true
      end
    end
  end
  return valid
end

local function testAchievements(json) -- checks and saves achievements
  local goals = jsonReadFile(event.achievementsFile)
  if goals then
    stats.achievementsDone = 0
    stats.achievementsTotal = json.achievements.total
    for k, v in pairs(goals) do
      if not tableContains(json.achievements.done, k) then
        if getCondition(v.cond) then
          table.remove(json.achievements.notDone, arrayFindValueIndex(json.achievements.notDone, k))
          table.insert(json.achievements.done, k)
          local data = {}
          data.image = "achievement"
          data.time = 8
          data.subtitle = v.title
          data.text = {v.desc}
          core_infoBox.queueInfo(data)
        end
      end
    end
    stats.achievementsDone = #json.achievements.done
  end
end

local function createRacePath() -- creates the entire race path, as well as the modified AI path
  if not mapNodes then mapNodes = map.getMap().nodes end
  if not mapGp then mapGp = map.getGraphpath() end
	local n1, n2 = nil, nil
  event.lapCountAi = event.lapCount > 1 and event.lapCount + 1 or 1
  event.waypointsAi = deepcopy(event.waypoints)
  local num = #event.waypoints
  n1 = map.findClosestRoad(vec3(scenetree.findObject(event.waypoints[num - 1]):getPosition()))
  n2 = map.findClosestRoad(vec3(scenetree.findObject(event.waypoints[num]):getPosition()))
  local shortPath = mapGp:getPath(n1, n2, 1)
  num = #shortPath
  finishVec = (mapNodes[shortPath[num]].pos - mapNodes[shortPath[num - 1]].pos):z0():normalized()
  if event.lapCount == 1 then
    local extendPath = mapGp:getRandomPathG(n2, finishVec, 200)
    table.insert(event.waypointsAi, extendPath[#extendPath])
  else
    table.insert(event.waypointsAi, event.waypoints[1])
  end
  n1, n2 = map.findClosestRoad(player.pos)
  if (mapNodes[n1].pos - player.pos):normalized():dot(player.vec) > 0 then
    n1, n2 = n2, n1
  end
  racePath = {{n1, 0}}
	for i, v in ipairs(event.waypoints) do
		n2 = map.findClosestRoad(vec3(scenetree.findObject(v):getPosition()))
		local shortPath = mapGp:getPath(n1, n2, 1)
    if not shortPath[1] then
      log("E", logTag, "Race path error! No path from "..n1.." to "..n2)
    end
		for o, w in ipairs(shortPath) do
			if o > 1 then
        raceLength = raceLength + mapNodes[shortPath[o - 1]].pos:distance(mapNodes[w].pos)
				table.insert(racePath, {w, raceLength})
			end
		end
		n1 = n2
	end
  raceLength = raceLength * event.lapCount
end

local function setupPlayer() -- sets up player
  player.obj = be:getPlayerVehicle(0)
	player.name = player.obj.name
	player.id = player.obj:getID()
  player.pos = vec3(player.obj:getPosition())
  player.vec = vec3(player.obj:getDirectionVector())
  player.vecUp = vec3()
  player.role = "player"
  if event.playerName then
    core_vehicles.setPlateText(event.playerName, player.id)
    player.driver = event.playerName
  end
end

local function setupRacers() -- sets up racers
  local quantity = spawn.group and spawn.quantity or 9
	local others = extensions.core_multiSpawn.getNewVehicles(quantity)
	for i, v in ipairs(others) do
		local r = {}
		r.obj = scenetree.findObjectById(v)
		r.name = r.obj.name
		r.id = v
    r.pos = vec3(r.obj:getPosition())
    r.vec = vec3()
    r.vecUp = vec3()
    r.role = "ai"
    if aiLua ~= "ai" then
      r.obj:queueLuaCommand("extensions.load('"..aiLua.."')")
    end
		table.insert(racers, r)
	end
end

local function setAggroScale(value) -- sets the base aggro scale (pseudo difficulty)
  if not value then value = 1 end
  aggroScale = value
end

local function adjustAggro(r) -- dynamically adjusts the aggression value of an AI racer, to help improve its ability
  if player.firstWaypoint and event.aggroRubberband > 0 then
    local rb = event.aggroRubberbandBounds
    r.aggro = clamp(r.aggroBase + clamp((player.path.percent - r.path.percent) / (15 / event.aggroRubberband), rb[1], rb[2]), 0.25, 1.25)
  end
  if event.watchNodes then
    for k, v in pairs(event.watchNodes) do
      if racePath[r.path.idx + 1] and k == racePath[r.path.idx + 1][1] then
        local nodesAhead = v.nodesAhead or 1
        r.aggroPathIdx = r.path.idx + v.nodesAhead + 1
        r.aggroFactor = 0
        r.aggroAlt = clamp(r.aggro * v.aggroValue, 0.25, 1.25)
        r.obj:queueLuaCommand(aiLua.."setAggression("..tostring(r.aggroAlt)..")")
        break
      end
    end
  end
  if r.aggroFactor > 0 and event.aggroTurnWatch > 0 and event.aggroTurnFactor > 0 and racePath[r.path.idx + 14] then -- look ahead at race nodes
    local newAggro = 1
    local newIdx
    local newSpeed = math.pow(r.speed, 1.2)
    local bounds = {square(30 + newSpeed), square(60 + newSpeed)}
    local pSet = {}
    for i = 1, 14 do
      pSet[i] = mapNodes[racePath[r.path.idx + i][1]].pos
    end
    for i = 1, 7 do
      local nodeDist = r.pos:squaredDistance(pSet[i])
      if nodeDist >= bounds[1] and nodeDist <= bounds[2] then
        newIdx = i
        break
      end
    end
    if newIdx then
      local height, point, xnorm
      local nDist, cDist = 0, 0
      local n = newIdx
      local dist = pSet[n]:squaredDistance(pSet[n + 1])
      local firstVec = (pSet[n + 1] - pSet[n]):z0():normalized()
      for i = n + 2, n + 7 do
        nDist = pSet[i - 1]:squaredDistance(pSet[i])
        dist = dist + nDist
        if dist >= square(80) then
          cDist = math.sqrt(dist) - 80
          xnorm = cDist / math.sqrt(nDist)
          point = linePointFromXnorm(pSet[i], pSet[i - 1], xnorm)
          height = math.max(0, pSet[1].z - pSet[i].z)
          break
        end
      end
      if not point then point = pSet[n + 7] end
      if not height then height = math.max(0, pSet[1].z - pSet[n + 7].z) end
      local dotValue = firstVec:dot((point - pSet[1]):z0():normalized())
      if dotValue <= event.aggroTurnWatch then
        dotValue = dotValue - (height * (event.aggroDownhill + newSpeed / 240)) / (math.sqrt(dist) - cDist)
        newAggro = clamp(r.aggro * clamp(dotValue * event.aggroTurnFactor, event.aggroTurnCutoff, 1), 0.25, 1.25)
        if dotValue < r.aggroFactor then
          r.aggroPathIdx = r.path.idx + newIdx + 1
          r.aggroFactor = dotValue
          r.aggroAlt = newAggro
          r.obj:queueLuaCommand(aiLua.."setAggression("..tostring(newAggro)..")")
        end
      end
    end
  end
  if r.path.idx >= r.aggroPathIdx then
    r.aggroAlt = math.min(r.aggro, r.aggroAlt + event.aggroReboundRate)
    r.obj:queueLuaCommand(aiLua.."setAggression("..tostring(r.aggroAlt)..")")
    r.aggroFactor = 1
  end
end

local function trackPlace() -- tracks the player standings throughout the race
  local place = 1
  for i, r in ipairs(racers) do
    if r.path.percent > player.path.percent then
      place = place + 1
    end
  end
  player.place = place
end

local function trackProgress(r) -- tracks the race path progress of the given vehicle
  local bestIdx = 0
  local bestDist = math.huge
  local newNode = nil
  local nodeCount = #racePath
  for i = math.max(1, r.path.idx - 5), math.min(nodeCount - 1, r.path.idx + 5) do
    local roadDist = r.pos:squaredDistanceToLineSegment(mapNodes[racePath[i][1]].pos, mapNodes[racePath[i + 1][1]].pos)
    if roadDist < bestDist then
      bestIdx = i
      bestDist = roadDist
    end
  end
  local idx = bestIdx
  local p1, p2 = mapNodes[racePath[idx][1]].pos, mapNodes[racePath[idx + 1][1]].pos
  local offset = linePointFromXnorm(p1, p2, clamp(p1:distance(r.pos) / p1:distance(p2), 0, 1))
  offset = offset:distance(p1)
  local percent = ((racePath[idx][2] + offset + racePath[nodeCount][2] * r.lapsDone) / raceLength) * 100
  newNode = {node = racePath[idx][1], idx = idx, percent = percent}
  r.path = newNode
end

local function trackStats(tick) -- tracks race stats
  local blockState = false
  stats.damage = player.damage
  if player.speed > stats.topSpeed and player.speed <= 100 then
    stats.topSpeed = player.speed
  end
  if stats.topSpeed <= event.crawlSpeed then
    stats.headStart = mainTimer
  end
  if player.vecUpDot < stats.maxRoll then
    stats.maxRoll = player.vecUpDot
  end
  if player.place == 1 then
    stats.leadTime = stats.leadTime + tick
    if stats.leadTime > stats.maxLeadTime then
      stats.maxLeadTime = stats.leadTime
    end
  else
    stats.leadTime = 0
  end
  if stats.halfwayPlace == 0 and player.path.percent >= 50 then
    stats.halfwayPlace = player.place
  end
  for i, r in ipairs(racers) do
    if r.state == "race" then
      if r.dist <= 8 and r.path.percent < player.path.percent and (player.pos - r.pos):normalized():dot(player.vec) >= 0.95 then
        blockState = true
      end
      if r.speed <= event.crawlSpeed then
        if not tableContains(stats.stuckRacers, r.id) then
          table.insert(stats.stuckRacers, r.id)
        end
      else
        if tableContains(stats.stuckRacers, r.id) then
          table.remove(stats.stuckRacers, arrayFindValueIndex(stats.stuckRacers, r.id))
        end
      end
      if r.vecUp:dot(up) <= 0 and r.dist <= 120 then
        if player.vec:dot((r.pos - player.pos):normalized()) > 0 then
          if not tableContains(stats.rolledRacers, r.id) then
            table.insert(stats.rolledRacers, r.id)
          end
        end
      end
    end
  end
  if blockState then
    stats.blockTime = stats.blockTime + tick
    if stats.blockTime > stats.maxBlockTime then
      stats.maxBlockTime = stats.blockTime
    end
  else
    stats.blockTime = 0
  end
end

local function insertNewPaths(data) -- inserts custom road segments into the map and reloads the map
  local mapFile = event.level.."map.json"
  local jsonMain = jsonReadFile(mapFile)
  if jsonMain then
    if not data.segments then
      data.segments = {}
    end
    if data.fileName then
      local json = jsonReadFile(data.fileName)
      if json then
        for k, v in pairs(json.segments) do
          data.segments[k] = v
        end
      end
    end
    for k, v in pairs(data.segments) do
      jsonMain.segments[k] = v
    end
    jsonWriteFile(event.level.."map.json", jsonMain, true)
    map.reset()
  end
end

local function spawnGroup() -- spawns a vehicle group
  if spawn.minQuantity then
    carLimit = carLimit - 1
    if userCarLimit and userCarLimit > 0 then
      carLimit = userCarLimit
    end
    spawn.quantity = math.min(9, clamp(carLimit, spawn.minQuantity, spawn.quantity))
  end
  core_multiSpawn.deleteVehicles(spawn.quantity)
  core_multiSpawn.setGroup("race", spawn.group)
  core_multiSpawn.setupVehicles(spawn.quantity, spawn.order, spawn.mode, spawn.gap)
end

local function initScenario(data) -- initializes the scenario; gets called by onScenarioLoaded
  extensions.load("core_infoBox")
  local this = data or scenario.getScenario()
  event.key = this.scenarioName
  event.level = path.split(getMissionFilename(), "(.-)([^/]-([^%.]*))$")
  event.waypoints = this.lapConfig
  event.lapCount = this.lapCount or math.huge
  event.lapPoint = event.waypoints[#event.waypoints]
  if this.manualPaths then
    insertNewPaths(this.manualPaths)
  end
  if this.spawn then
    spawn = this.spawn
  end
  if this.raceExtras then
    for k, v in pairs(this.raceExtras) do
      if type(v) ~= "nil" then
        event[k] = v
      end
    end
  end
  event.watchNodes = this.watchNodes
  if event.seriesFile then
    series = jsonReadFile(event.seriesFile)
  end
  if series.infoBoxStyles then
    core_infoBox.setStyles(series.infoBoxStyles)
  end
  event.messages = series.messages or event.messages
  event.saveFile = series.saveFile
  event.achievementsFile = series.achievementsFile
  if not data then
    if spawn.group then
      spawnDelay = 0
    end
  end
end

local function onScenarioUIReady()
  local json = jsonReadFile(event.saveFile)
  stats = {topSpeed = 0, damage = 0, maxRoll = 1, headStart = 0, leadTime = 0, maxLeadTime = 0, blockTime = 0, maxBlockTime = 0, racersCount = 0, finishTime = 0, finishPlace = 0, finishGap = math.huge, finishDrive = 1, secrets = 0, halfwayPlace = 0, reboundDifference = 0, stuckRacers = {}, rolledRacers = {}}
  mainTimer = 0
  if firstLoad then
    if not json or json.series ~= series.name then
      json = setupSaveFile()
    end
    aiLua = event.aiLua
    setupPlayer()
    setupRacers()
    aiLua = aiLua.."."
    finishPos = vec3(scenetree.findObject(event.waypoints[#event.waypoints]):getPosition())
    finishRadius = getSceneWaypointRadius(scenetree.findObject(event.waypoints[#event.waypoints]))
    createRacePath()
    firstLoad = false
  end
  player.firstWaypoint = false
  for i, r in ipairs(racers) do
    r.obj:queueLuaCommand(aiLua.."setMode('stop')")
    r.obj:queueLuaCommand(aiLua.."setAggression(1.4)")
    r.obj:queueLuaCommand("controller.setFreeze(1)")
    r.obj:resetBrokenFlexMesh()
    if event.racerNames[i] then
      core_vehicles.setPlateText(event.racerNames[i], r.id)
      r.driver = event.racerNames[i]
    end
  end
  stats.racersCount = #racers
  if event.useMainBox then
    local data = json
    data.image = "info"
    data.racesList = series.races
    data.difficultyList = {"Easy", "Normal"}
    data.colorsList = {}
    data.playerColors = core_multiSpawn.getColors(player.obj.jbeam, true)
    if series.infoBoxStyles and series.infoBoxStyles.text then
      data.title = series.infoBoxStyles.text.title
      data.subtitle = series.infoBoxStyles.text.subtitle
    end
    for k, v in pairs(data.playerColors) do
      table.insert(data.colorsList, k)
    end
    table.sort(data.colorsList)
    if json.useAchievements then
      data.achievementsDoneList, data.achievementsNotDoneList = {}, {}
      data.achievementsRef = jsonReadFile(series.achievementsFile)
      for i, v in ipairs(json.achievements.done) do
        table.insert(data.achievementsDoneList, v)
      end
      table.sort(data.achievementsDoneList)
      for i, v in ipairs(json.achievements.notDone) do
        table.insert(data.achievementsNotDoneList, v)
      end
      table.sort(data.achievementsNotDoneList)
    end
    core_infoBox.setupMainBox(data, "mini")
  end
end

local function onCountdownStarted()
  player.path = {node = "", idx = 0, percent = 0}
	player.timers = {stuck = 0, flip = 0}
	player.lapsDone = 0
  player.place = 0
  player.state = "start"
  finishers = {}
  scoreStats.initialiseArbitraryStat("score", "Score", player.obj, player.obj.name, 5, 5)
  for i, r in ipairs(racers) do
    r.aggroBase = event.aggroValue * aggroScale + (math.random() * 2 - 1) * event.aggroVariance
    r.aggro = r.aggroBase
		r.path = {node = "", idx = 0, percent = 0}
		r.lapsDone = 0
    r.aggroPathIdx = 3
    r.aggroFactor = 1
    r.aggroAlt = r.aggro
    r.distTarget = 0
		r.state = "start"
    if event.forceLateralValue then
      r.forceLateral = event.forceLateralVariance and event.forceLateralValue + (math.random() * 2 - 1) * event.forceLateralVariance or event.forceLateralValue
      r.obj:queueLuaCommand(aiLua.."setForceLateral("..tostring(r.forceLateral)..")")
    end
    if event.distPastEdgeValue then
      r.distPastEdge = event.distPastEdgeVariance and event.distPastEdgeValue + (math.random() * 2 - 1) * event.distPastEdgeVariance or event.distPastEdgeValue
      r.obj:queueLuaCommand(aiLua.."setDistPastEdge("..tostring(r.distPastEdge)..")")
    end
    r.obj:queueLuaCommand("controller.getControllerSafe('esc').setESCMode(2)")
    r.obj:queueLuaCommand(aiLua.."driveUsingPath({wpTargetList = "..serialize(event.waypointsAi)..", noOfLaps = "..tostring(event.lapCountAi)..", aggression = "..tostring(math.min(1.4, r.aggro * 1.2))..", avoidCars = "..tostring(event.avoidCars).."})")
  end
  if event.useMainBox then
    core_infoBox.setupMainBox(nil, "hidden")
  end
end

local function onRaceStart()
  player.state = "race"
  for i, r in ipairs(racers) do
    r.obj:queueLuaCommand("controller.setFreeze(0)")
    r.state = "race"
    if event.debug then
      r.obj:queueLuaCommand(aiLua.."setVehicleDebugMode({debugMode = 'speeds'})")
    end
  end
end

local function onRaceTick(tick)
	player.vec:set(player.obj:getDirectionVector())
  player.vecUp:set(player.obj:getDirectionVectorUp())
  player.vecUpDot = player.vecUp:dot(up)
	player.damage = map.objects[player.id].damage or 0
  player.speed = map.objects[player.id].vel:length()
  if player.timers.flip >= event.stuckTime then
    scenario.finish({failed = event.messages.flip})
  end
  if player.timers.stuck >= event.stuckTime then
    scenario.finish({failed = event.messages.stuck})
  end
  if player.speed <= event.crawlSpeed then
    if player.vecUpDot <= 0.1 then
      player.timers.flip = player.timers.flip + tick
    else
      player.timers.flip = 0
    end
    if player.timers.flip == 0 and player.damage >= event.damageCutoff then
      player.timers.stuck = player.timers.stuck + tick
    end
  else
    player.timers.stuck = 0
  end
	for i, r in ipairs(racers) do
		r.vec:set(r.obj:getDirectionVector())
    r.vecUp:set(r.obj:getDirectionVectorUp())
		r.damage = map.objects[r.id].damage
    r.speed = map.objects[r.id].vel:length()
    r.dist = r.pos:distance(player.pos)
    trackProgress(r)
    if r.state == "race" then
      adjustAggro(r)
    end
	end
  trackProgress(player)
  trackPlace()
  trackStats(tick)
  mainTimer = mainTimer + tick
end

local function onRaceResult()
  if player.state == "finish" then
    stats.finishTime = scenario.getScenario().timer
    local finishText = ""
    local errorText = "Error: Invalid race stats!"
    local reward, medalScore, raceScore = 0, 0, 0
    for i, v in ipairs(finishers) do
      if v == "Player" then
        stats.finishPlace = i
        medalScore = math.max(1, 6 - i)
        raceScore = math.max(1, 11 - i)
      end
    end
    scoreStats.setStatProgress(player.obj:getID(), "score", player.obj.name, {value = medalScore})
    if finishers[1] then
      if finishers[1] == "Player" then
        finishText = event.messages.win
      else
        finishText = finishers[1].." "..event.messages.finish
        finishText = finishText.." "..event.messages.place.." "..ranks[stats.finishPlace].."."
      end
      if event.rewards[stats.finishPlace] then
        local multiplier = aggroScale == 1 and 1 or 0.5
        reward = math.ceil(event.rewards[stats.finishPlace] * multiplier)
      end
      local json = jsonReadFile(event.saveFile)
      local race = json.races[event.key]
      json.money = json.money + reward
      json.finishes = json.finishes + 1
      race.score = math.max(raceScore, race.score)
      race.finishes = race.finishes + 1
      if race.time == 0 or stats.finishTime <= race.time then
        race.time = stats.finishTime
      end
      if raceScore == 10 then
        json.wins = json.wins + 1
        race.wins = race.wins + 1
      else
        stats.headStart = 0
      end
      stats.reboundDifference = stats.halfwayPlace - stats.finishPlace
      stats.money = json.money
      stats.wins = json.wins
      stats.totalScore = 0
      stats.minScore = 10
      stats.seriesTime = 0
      for k, v in pairs(json.races) do
        stats.totalScore = stats.totalScore + v.score
        if v.score < stats.minScore then
          stats.minScore = v.score
        end
        json.minScore = stats.minScore
        if stats.seriesTime < 1e6 and v.score > 0 then
          stats.seriesTime = stats.seriesTime + v.time
        else
          stats.seriesTime = 1e6
        end
      end
      json.time = stats.seriesTime
      if json.useAchievements then
        stats.stuckRacersCount = #stats.stuckRacers
        stats.rolledRacersCount = #stats.rolledRacers
        testAchievements(json)
      end
      finishText = finishText.." "..event.messages.finishTime.." "..timeToString(stats.finishTime).."."
      finishText = finishText.." "..event.messages.finishReward..reward.."."
      finishText = finishText.." "..event.messages.money..json.money.."."
      if json.useAchievements then
        finishText = finishText.." "..event.messages.achievements.." "..stats.achievementsDone.."/"..stats.achievementsTotal.." ."
        if stats.achievementsDone >= stats.achievementsTotal then
          finishText = finishText.." "..event.messages.congrats
        end
      end
      if json.money >= 1000000 or json.money < 0 then
        finishText = finishText.." It looks like you cheated with your money. Yoink, it's all gone now!"
        json.money = 0
      end
      jsonWriteFile(event.saveFile, json)
    end
    if stats.finishPlace == 0 then
      scenario.finish({failed = errorText})
    elseif stats.finishPlace <= 3 then
      scenario.finish({msg = finishText})
    else
      scenario.finish({failed = finishText})
    end
  end
end

local function onRaceWaypointReached(data)
  if data.vehicleId == player.id then
    if data.waypointName == event.waypoints[1] then
      player.firstWaypoint = true
    elseif data.waypointName == event.lapPoint then
      player.lapsDone = player.lapsDone + 1
      player.path.idx = 1
      if player.lapsDone >= event.lapCount then
        table.insert(finishers, "Player")
        player.state = "finish"
        if finishers[1] == "Player" then
          local bestDist = math.huge
          for i, r in ipairs(racers) do
            if r.dist < bestDist then
              bestDist = r.dist
            end
          end
          stats.finishGap = bestDist
        end
        stats.finishDrive = player.vec:dot(finishVec)
      end
    end
  end
end

local function onBeamNGTrigger(data)
  if data.subjectName == player.name and data.event == "enter" then
    if string.find(data.triggerName, "secret") then
      Engine.Audio.playOnce("AudioGui", "sounds/gotSecret.wav")
      stats.secrets = stats.secrets + 1
    end
  end
end

local function onPreRender(dt)
  local current = be:getPlayerVehicle(0)
  if not current or current:getID() ~= playerObj:getID() then
    be:enterVehicle(0, playerObj)
  end
  if spawnDelay == 2 then
    spawnGroup()
    spawnDelay = -1
  elseif spawnDelay >= 0 then
    spawnDelay = spawnDelay + 1
  end
  if #racers > 0 then
    player.pos:set(player.obj:getPosition())
    for i, r in ipairs(racers) do
      if event.aiLua ~= "ai" then
        r.obj:queueLuaCommand(aiLua.."updateGFX("..tostring(dt)..")")
      end
      r.pos:set(r.obj:getPosition())
      r.distTarget = r.pos:squaredDistance(finishPos)
      if r.state == "race" and r.distTarget <= square(finishRadius) then
        if r.path.idx >= 3 then
          r.lapsDone = r.lapsDone + 1
          r.path.idx = 1
        end
        if r.lapsDone >= event.lapCount then
          r.obj:queueLuaCommand(aiLua.."setAggression(0.25)")
          local str = r.driver or "Racer #"..r.id
          table.insert(finishers, str)
          r.state = "finish"
        end
      end
    end
  end
end

local function onRaceExtrasAction(data) -- custom race data actions
  for k, v in pairs(data) do
    if k == "playerColor" then
      setCarColor(player.id, v)
    end
    if k == "difficulty" then
      local value = v == "Easy" and 0.75 or 1
      setAggroScale(value)
    end
    if k == "resetS" then
      modifySaveFile({stats = true})
    end
    if k == "resetA" then
      modifySaveFile({achievements = true})
    end
  end
end

local function debugEventValue(key, value) -- sets event values manually
  event[key] = value
end

local function onExtensionUnloaded()
  os.remove(event.level.."map.json")
end

local function onExit()
  os.remove(event.level.."map.json")
end

local function onSerialize()
  local data = scenario.getScenario()
  return data
end

local function onDeserialized(data)
  initScenario(data)
end

M.initScenario = initScenario
M.timeToString = timeToString
M.setupSaveFile = setupSaveFile
M.modifySaveFile = modifySaveFile
M.debugEventValue = debugEventValue
M.onScenarioUIReady = onScenarioUIReady
M.onCountdownStarted = onCountdownStarted
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onRaceWaypointReached = onRaceWaypointReached
M.onBeamNGTrigger = onBeamNGTrigger
M.onPreRender = onPreRender
M.onRaceExtrasAction = onRaceExtrasAction
M.onSerialize = onSerialize
M.onDeserialized = onDeserialized
M.onExtensionUnloaded = onExtensionUnloaded
M.onExit = onExit

return M